"""PyGNA2 is based on `PyGNA <https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-020-03801-1>`_. It uses the same statistical tests, but improves the way results are reported.

PyGNA is a tool that can perform several kinds of statistical tests on gene networks and gene sets.

- Publication: https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-020-03801-1
- Documentation: https://pygna.readthedocs.io/en/latest/index.html
- Github: https://github.com/stracquadaniolab/pygna
- Workflow github: https://github.com/stracquadaniolab/workflow-pygna
"""